"""
Swarm Communication Protocol
- RAG as a communication protocol
- Each Agent is connected to a database so they can see each others 
memories, actions, and experiences

"""

import json
from swarms.structs.omni_agent_types import AgentType
from swarms.structs.base_structure import BaseStructure
from typing import List
from swarms.memory.base_vectordb import BaseVectorDatabase
import time
from swarms.utils.loguru_logger import logger
from pydantic import BaseModel, Field


class SwarmCommunicationProtocol(BaseModel):
    agent_name: str = Field(
        None, title="Agent Name", description="The name of the agent"
    )
    message: str = Field(
        None, title="Message", description="The message to be sent"
    )
    timestamp: float = Field(
        None,
        title="Timestamp",
        description="The time the message was sent",
    )


class SCP(BaseStructure):
    def __init__(
        self,
        agents: List[AgentType],
        memory_system: BaseVectorDatabase = None,
        *args,
        **kwargs,
    ):
        super().__init__()
        self.agents = agents
        self.memory_system = memory_system

        # Memory system
        if memory_system is not None:
            for agent in self.agents.values():
                agent.long_term_memory = memory_system

        logger.info(
            "AgentRearrange initialized with agents: {}".format(
                list(self.agents.keys())
            )
        )

    def message_log(
        self, agent: AgentType, task: str = None, message: str = None
    ):
        log = {
            "agent_name": agent.agent_name,
            "task": task,
            "response": message,
            "timestamp": time.time(),
        }

        # Transform the log into a string
        log_output = json.dumps(log)

        # Add the log to the memory system
        self.memory_system.add(log)

        return log_output

    def run_single_agent(
        self, agent: AgentType, task: str, *args, **kwargs
    ):
        # Send the message to the agent
        output = agent.run(task)

        # log the message
        self.message_log(
            agent=agent,
            task=task,
            message=output,
        )

        # Log the output
        return output

    def send_message(self, agent: AgentType, message: str):
        agent.receieve_mesage(self.message_log(agent, message))



import json
from typing import List

# from beartype import beartype

from swarms.structs.agent import Agent
from swarms.structs.base_swarm import BaseSwarm
from swarms.utils.loguru_logger import logger
from pydantic import BaseModel, Field
from swarms.structs.conversation import Conversation


class HiearchicalRequest(BaseModel):
    task: str = Field(
        None,
        title="Task",
        description="The task to send to the director agent.",
    )
    agent_name: str = Field(
        None,
        title="Agent Name",
        description="The name of the agent to send the task to.",
    )


class HiearchicalRequestDict(BaseModel):
    task: str = Field(
        None,
        title="Task",
        description="The task to send to the director agent.",
    )
    agent_name: str = Field(
        None,
        title="Agent Name",
        description="The name of the agent to send the task to.",
    )

    class Config:
        schema_extra = {
            "example": {
                "task": "task",
                "agent_name": "agent_name",
            }
        }


"""
Boss -> json -> workers -> json -> Boss


Parse the JSON data and activate the selected agent.

parse -> execute
"""


class HiearchicalSwarm(BaseSwarm):
    """
    A class representing a hierarchical swarm.

    Attributes:
        name (str): The name of the hierarchical swarm.
        description (str): The description of the hierarchical swarm.
        director (Agent): The director agent of the hierarchical swarm.
        agents (List[Agent]): The list of agents in the hierarchical swarm.
        max_loops (int): The maximum number of loops to run the swarm.
        long_term_memory_system (BaseSwarm): The long term memory system of the swarm.
        custom_parse_function (callable): A custom parse function for the swarm.

    Methods:
        swarm_initialization(*args, **kwargs): Initializes the hierarchical swarm.
        find_agent_by_name(agent_name: str = None, *args, **kwargs): Finds an agent in the swarm by name.
        parse_function_activate_agent(json_data: str = None, *args, **kwargs): Parses JSON data and activates the selected agent.
        select_agent_and_send_task(name: str = None, task: str = None, *args, **kwargs): Selects an agent and sends a task to them.
        run(task: str = None, *args, **kwargs): Runs the hierarchical swarm.

    """

    def __init__(
        self,
        name: str = None,
        description: str = None,
        director: Agent = None,
        agents: List[Agent] = None,
        max_loops: int = 1,
        long_term_memory_system: BaseSwarm = None,
        custom_parse_function: callable = None,
        rules: str = None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.name = name
        self.description = description
        self.director = director
        self.agents = agents
        self.max_loops = max_loops
        self.long_term_memory_system = long_term_memory_system
        self.custom_parse_function = custom_parse_function

        # Set the director to max_one loop
        if self.director.max_loops > 1:
            self.director.max_loops = 1

        # Set the long term memory system of every agent to long term memory system
        if long_term_memory_system is True:
            for agent in agents:
                agent.long_term_memory = long_term_memory_system

        # Initialize the swarm
        self.swarm_initialization()

        # Initialize the conversation message pool
        self.swarm_history = Conversation(
            time_enabled=True,
        )

    def swarm_initialization(self, *args, **kwargs):
        """
        Initializes the hierarchical swarm.

        Args:
            *args: Additional positional arguments.
            **kwargs: Additional keyword arguments.

        Returns:
            None

        """
        logger.info(f"Initializing the hierarchical swarm: {self.name}")
        logger.info(f"Purpose of this swarm: {self.description}")

        # Now log number of agnets and their names
        logger.info(f"Number of agents: {len(self.agents)}")
        logger.info(
            f"Agent names: {[agent.name for agent in self.agents]}"
        )

        # Now see if agents is not empty
        if len(self.agents) == 0:
            logger.info("No agents found. Please add agents to the swarm.")
            return None

        # Now see if director is not empty
        if self.director is None:
            logger.info(
                "No director found. Please add a director to the swarm."
            )
            return None

        logger.info(
            f"Initialization complete for the hierarchical swarm: {self.name}"
        )

    def find_agent_by_name(self, agent_name: str = None, *args, **kwargs):
        """
        Finds an agent in the swarm by name.

        Args:
            agent_name (str): The name of the agent to find.

        Returns:
            Agent: The agent with the specified name, or None if not found.

        """
        for agent in self.agents:
            if agent.name == agent_name:
                return agent
        return None

    def parse_function_activate_agent(
        self, json_data: str = None, *args, **kwargs
    ):
        """
        Parse the JSON data and activate the selected agent.

        Args:
            json_data (str): The JSON data containing the agent name and task.

        Returns:
            str: The response from the activated agent.

        Raises:
            json.JSONDecodeError: If the JSON data is invalid.

        """
        try:
            data = json.loads(json_data)

            # Check if the data is a list of agent task pairs
            if isinstance(data, list):
                responses = []
                # Iterate over the list of agent task pairs
                for agent_task in data:
                    name = agent_task.get("name")
                    task = agent_task.get("task")

                    response = self.select_agent_and_send_task(
                        name, task, *args, **kwargs
                    )

                    responses.append(response)
                return responses
            else:
                name = data.get("name")
                task = data.get("task")

                response = self.select_agent_and_send_task(
                    name, task, *args, **kwargs
                )

                return response
        except json.JSONDecodeError:
            logger.error("Invalid JSON data, try again.")
            raise json.JSONDecodeError

    def select_agent_and_send_task(
        self, name: str = None, task: str = None, *args, **kwargs
    ):
        """
        Select an agent from the list and send a task to them.

        Args:
            name (str): The name of the agent to send the task to.
            task (str): The task to send to the agent.

        Returns:
            str: The response from the agent.

        Raises:
            KeyError: If the agent name is not found in the list of agents.

        """
        try:
            # Check to see if the agent name is in the list of agents
            if name in self.agents:
                agent = self.agents[name]
            else:
                return "Invalid agent name. Please select 'Account Management Agent' or 'Product Support Agent'."

            response = agent.run(task, *args, **kwargs)

            return response
        except Exception as e:
            logger.error(f"Error: {e}")
            raise e

    def run(self, task: str = None, *args, **kwargs):
        """
        Run the hierarchical swarm.

        Args:
            task (str): The task to send to the director agent.

        Returns:
            str: The response from the director agent.

        Raises:
            Exception: If an error occurs while running the swarm.

        """
        try:
            loop = 0

            # While the loop is less than max loops
            while loop < self.max_loops:
                # Run the director
                response = self.director.run(task, *args, **kwargs)

                # Log the director's response
                self.swarm_history.add(self.director.agent_name, response)

                # Run agents
                if self.custom_parse_function is not None:
                    response = self.custom_parse_function(response)
                else:
                    response = self.parse_function_activate_agent(response)

                loop += 1

                task = response

            return response
        except Exception as e:
            logger.error(f"Error: {e}")
            raise e



import json
import random
from typing import List

import tenacity
from swarms.structs.agent import Agent
from swarms.utils.logger import logger
from swarms.structs.base_swarm import BaseSwarm


# [TODO]: Add type hints
class MultiAgentCollaboration(BaseSwarm):
    """
    Multi-agent collaboration class.

    Attributes:
        agents (List[Agent]): The agents in the collaboration.
        selection_function (callable): The function that selects the next speaker.
            Defaults to select_next_speaker.
        max_iters (int): The maximum number of iterations. Defaults to 10.
        autosave (bool): Whether to autosave the state of all agents. Defaults to True.
        saved_file_path_name (str): The path to the saved file. Defaults to
            "multi_agent_collab.json".
        stopping_token (str): The token that stops the collaboration. Defaults to
            "<DONE>".
        results (list): The results of the collaboration. Defaults to [].
        logger (logging.Logger): The logger. Defaults to logger.
        logging (bool): Whether to log the collaboration. Defaults to True.


    Methods:
        reset: Resets the state of all agents.
        inject: Injects a message into the collaboration.
        inject_agent: Injects an agent into the collaboration.
        step: Steps through the collaboration.
        ask_for_bid: Asks an agent for a bid.
        select_next_speaker: Selects the next speaker.
        run: Runs the collaboration.
        format_results: Formats the results of the run method.


    Usage:
    >>> from swarms.models import OpenAIChat
    >>> from swarms.structs import Agent
    >>> from swarms.swarms.multi_agent_collab import MultiAgentCollaboration
    >>>
    >>> # Initialize the language model
    >>> llm = OpenAIChat(
    >>>     temperature=0.5,
    >>> )
    >>>
    >>>
    >>> ## Initialize the workflow
    >>> agent = Agent(llm=llm, max_loops=1, dashboard=True)
    >>>
    >>> # Run the workflow on a task
    >>> out = agent.run("Generate a 10,000 word blog on health and wellness.")
    >>>
    >>> # Initialize the multi-agent collaboration
    >>> swarm = MultiAgentCollaboration(
    >>>     agents=[agent],
    >>>     max_iters=4,
    >>> )
    >>>
    >>> # Run the multi-agent collaboration
    >>> swarm.run()
    >>>
    >>> # Format the results of the multi-agent collaboration
    >>> swarm.format_results(swarm.results)

    """

    def __init__(
        self,
        name: str = "MultiAgentCollaboration",
        description: str = "A multi-agent collaboration.",
        director: Agent = None,
        agents: List[Agent] = None,
        select_next_speaker: callable = None,
        max_iters: int = 10,
        autosave: bool = True,
        saved_file_path_name: str = "multi_agent_collab.json",
        stopping_token: str = "<DONE>",
        logging: bool = True,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.name = name
        self.description = description
        self.director = director
        self.agents = agents
        self.select_next_speaker = select_next_speaker
        self._step = 0
        self.max_iters = max_iters
        self.autosave = autosave
        self.saved_file_path_name = saved_file_path_name
        self.stopping_token = stopping_token
        self.results = []
        self.logger = logger
        self.logging = logging

    def inject(self, name: str, message: str):
        """Injects a message into the multi-agent collaboration."""
        for agent in self.agents:
            agent.run(f"Name {name} and message: {message}")
        self._step += 1

    def step(self) -> tuple[str, str]:
        """Steps through the multi-agent collaboration."""
        speaker_idx = self.select_next_speaker(self._step, self.agents)
        speaker = self.agents[speaker_idx]
        message = speaker.send()

        for receiver in self.agents:
            receiver.receive(speaker.name, message)
        self._step += 1

        if self.logging:
            self.log_step(speaker, message)

        return speaker.name, message

    def log_step(self, speaker: str, response: str):
        """Logs the step of the multi-agent collaboration."""
        self.logger.info(f"{speaker.name}: {response}")

    @tenacity.retry(
        stop=tenacity.stop_after_attempt(10),
        wait=tenacity.wait_none(),
        retry=tenacity.retry_if_exception_type(ValueError),
        before_sleep=lambda retry_state: print(
            f"ValueError occured: {retry_state.outcome.exception()},"
            " retying..."
        ),
        retry_error_callback=lambda retry_state: 0,
    )
    def select_next_speaker_bid(
        self,
        step: int,
        agents: List[Agent],
    ) -> int:
        """Selects the next speaker."""
        bids = []
        for agent in agents:
            bid = self.ask_for_bid(agent)
            bids.append(bid)
        max_value = max(bids)
        max_indices = [i for i, x in enumerate(bids) if x == max_value]
        idx = random.choice(max_indices)
        return idx

    @tenacity.retry(
        stop=tenacity.stop_after_attempt(10),
        wait=tenacity.wait_none(),
        retry=tenacity.retry_if_exception_type(ValueError),
        before_sleep=lambda retry_state: print(
            f"ValueError occured: {retry_state.outcome.exception()},"
            " retying..."
        ),
        retry_error_callback=lambda retry_state: 0,
    )
    def run_director(self, task: str):
        """Runs the multi-agent collaboration."""
        n = 0
        self.reset()
        self.inject("Debate Moderator")
        print("(Debate Moderator): ")
        print("\n")

        while n < self.max_iters:
            name, message = self.step()
            print(f"({name}): {message}")
            print("\n")
            n += 1

    def select_next_speaker_roundtable(
        self, step: int, agents: List[Agent]
    ) -> int:
        """Selects the next speaker."""
        return step % len(agents)

    def select_next_speaker_director(
        step: int, agents: List[Agent], director
    ) -> int:
        # if the step if even => director
        # => director selects next speaker
        if step % 2 == 1:
            idx = 0
        else:
            idx = director.select_next_speaker() + 1
        return idx

    def run(self, task: str, *args, **kwargs):
        # [TODO]: Add type hints
        # [TODO]: Implement the run method using step method
        conversation = task

        for _ in range(self.max_iters):
            for agent in self.agents:
                result = agent.run(conversation, *args, **kwargs)
                self.results.append({"agent": agent, "response": result})
                conversation += result

                if self.autosave:
                    self.save_state()
                if result == self.stopping_token:
                    break

        return self.results

    def format_results(self, results):
        """Formats the results of the run method"""
        formatted_results = "\n".join(
            [
                f"{result['agent']} responded: {result['response']}"
                for result in results
            ]
        )
        return formatted_results

    def save(self):
        """Saves the state of all agents."""
        state = {
            "step": self._step,
            "results": [
                {"agent": r["agent"].name, "response": r["response"]}
                for r in self.results
            ],
        }

        with open(self.saved_file_path_name, "w") as file:
            json.dump(state, file)

    def load(self):
        """Loads the state of all agents."""
        with open(self.saved_file_path_name) as file:
            state = json.load(file)
        self._step = state["step"]
        self.results = state["results"]
        return state

    def __repr__(self):
        return (
            f"MultiAgentCollaboration(agents={self.agents},"
            f" selection_function={self.select_next_speaker},"
            f" max_iters={self.max_iters}, autosave={self.autosave},"
            f" saved_file_path_name={self.saved_file_path_name})"
        )
